// ---------------------------------------------------------------------------

#include <fmx.h>
#pragma hdrstop

#include "FServer.h"
// ---------------------------------------------------------------------------
#pragma package(smart_init)
#pragma resource "*.fmx"
TFMicroServer* FMicroServer;
// ---------------------------------------------------------------------------
class TNotifier : public TCppInterfacedObject<TThreadProcedure> {
	int lValue;
private:
	TFMicroServer* form;
public:
	TNotifier(int ValSet, TFMicroServer* _form) : lValue(ValSet), form(_form) {}
	INTFOBJECT_IMPL_IUNKNOWN(TInterfacedObject);
	void __fastcall Invoke() {form->clients_connected->Text = lValue;}
} ;
// ---------------------------------------------------------------------------

class TLogMessage : public TCppInterfacedObject<Classes::TThreadProcedure> {
	String Msg;
private:
	TFMicroServer* form;
public:
	TLogMessage(String _Msg, TFMicroServer* _form) : Msg(_Msg), form(_form) {}
	INTFOBJECT_IMPL_IUNKNOWN(TInterfacedObject);

	void __fastcall Invoke() {form->messagesLog->Lines->Add(Msg);}
} ;

// ---------------------------------------------------------------------------
__fastcall TFMicroServer::TFMicroServer(TComponent* Owner) : TForm(Owner) {
	IdTCPServer->MaxConnections = 20;

}
// ---------------------------------------------------------------------------
void __fastcall TFMicroServer::IdTCPServerConnect(TIdContext* AContext) {
	UnicodeString ip;
	int port;
	UnicodeString peerIP;
	int peerPort;
	int nClients;
	UnicodeString msgToClient;
	UnicodeString typeClient;

	// ... OnConnect is a TIdServerThreadEvent property that represents the event
	// handler signalled when a new client connection is connected to the server.

	// ... Use OnConnect to perform actions for the client after it is connected
	// and prior to execution in the OnExecute event handler.

	// ... see indy doc:
	// http://www.indyproject.org/sockets/docs/index.en.aspx

	// ... getting IP address and Port of Client that connected
	ip = AContext->Binding->IP;
	port = AContext->Binding->Port;
	peerIP = AContext->Binding->PeerIP;
	peerPort = AContext->Binding->PeerPort;
	UnicodeString Msg;
	Msg.sprintf(L"Port %s PeerIP = - %s PeerPort %s ", IntToStr(port), peerIP,
	  IntToStr(peerPort));
	Display(L"SERVER", L"Client Connected!");
	Display(L"SERVER", Msg);

	// ... display the number of clients connected
	ShowNumberOfClients(false);

	// ... CLIENT CONNECTED:
	if (port == GUEST_CLIENT_PORT)
		typeClient = "GUEST";

	// ... send the Welcome message to Client connected
	msgToClient = "Welcome " + typeClient + " " + "Client :)";
	AContext->Connection->IOHandler->WriteLn(msgToClient);
}
// ---------------------------------------------------------------------------
void __fastcall TFMicroServer::IdTCPServerDisconnect(TIdContext* AContext) {
	UnicodeString ip;
	int port;
	UnicodeString peerIP;
	int peerPort;
	int nClients;

	// ... getting IP address and Port of Client that connected
	ip = AContext->Binding->IP;
	port = AContext->Binding->Port;
	peerIP = AContext->Binding->PeerIP;
	peerPort = AContext->Binding->PeerPort;

	// ... message log
	UnicodeString Msg;
	Msg.sprintf(L"Client Disconnected! Peer= %s : %s", peerIP,
	  IntToStr(peerPort));
	Display("SERVER", Msg);

	// ... display the number of clients connected
	ShowNumberOfClients(true);

}
// ---------------------------------------------------------------------------
void __fastcall TFMicroServer::IdTCPServerExecute(TIdContext* AContext) {
	UnicodeString ip;
	UnicodeString peerIP;
	int peerPort;
	UnicodeString msgFromClient;
	UnicodeString msgToClient;
	// ... OnExecute is a TIdServerThreadEvents event handler used to execute
	// the task for a client connection to the server.

	// ... here you can check connection status and buffering before reading
	// messages from client

	// ... see doc:
	// ... AContext.Connection.IOHandler.InputBufferIsEmpty
	// ... AContext.Connection.IOHandler.CheckForDataOnSource(<milliseconds>);
	// (milliseconds to wait for the connection to become readable)
	// ... AContext.Connection.IOHandler.CheckForDisconnect;

	// ... received a message from the client

	// ... get message from client
	msgFromClient = AContext->Connection->IOHandler->ReadLn();

	// ... getting IP address, Port and PeerPort from Client that connected
	peerIP = AContext->Binding->PeerIP;
	peerPort = AContext->Binding->PeerPort;

	// ... message log
	UnicodeString Msg;
	Msg.sprintf(L"(Peer= %s : %s )%s", peerIP, IntToStr(peerPort),
	  msgFromClient);
	Display(L"CLIENT", Msg);

	// ...

	// ... process message from Client

	// ...

	// ... send response to Client

	AContext->Connection->IOHandler->WriteLn
	  (L"... message sent from server :)");

}
// ---------------------------------------------------------------------------
void __fastcall TFMicroServer::IdTCPServerStatus(TObject* ASender,
  const TIdStatus AStatus, const UnicodeString AStatusText) {
	// ... OnStatus is a TIdStatusEvent property that represents the event handler
	// triggered when the current connection state is changed...

	// ... message log
	Display("SERVER", AStatusText);
}
// ---------------------------------------------------------------------------
void __fastcall TFMicroServer::ShowNumberOfClients(bool p_disconnected) {
	int nClients;
	try {
		IdTCPServer->Contexts->LockList();
		nClients = IdTCPServer->Contexts->Count();
	}

	__finally {
		IdTCPServer->Contexts->UnlockList();
	} // ... get number of clients connected

	// ... client disconnected?
	if (p_disconnected)
		nClients--;

	TThread::Queue(NULL, Classes::_di_TThreadProcedure(new TNotifier(nClients,
	  this)));
}
// ---------------------------------------------------------------------------
void __fastcall TFMicroServer::BroadcastMessage(UnicodeString p_message) {
	TList* tmpList;
	TIdContext* contexClient;
	int nClients;
	// ... send a message to all clients connected

	// ... get context Locklist
	tmpList = IdTCPServer->Contexts->LockList();
	try {
		int i = 0;
		while (i < tmpList->Count) {
			// ... get context (thread of i-client)
			contexClient = (TIdContext*)&tmpList[i];

			// ... send message to client
			contexClient->Connection->IOHandler->WriteLn(p_message);
			i++;
		}
	} __finally {
		// ... unlock list of clients!
		IdTCPServer->Contexts->UnlockList();
	}

}
// ---------------------------------------------------------------------------
void __fastcall TFMicroServer::Display(UnicodeString p_sender,
  UnicodeString p_message) {
	UnicodeString Tmp("[" + p_sender + "]" + GetNow() + ":" + p_message);
	TThread::Queue(NULL, Classes::_di_TThreadProcedure(new TLogMessage(Tmp,
	  this)));
}
// ---------------------------------------------------------------------------
UnicodeString __fastcall TFMicroServer::GetNow() {
	return FormatDateTime(L"dd.mm.yyyy - hh:nn:s:z", Now()) + "': ";}
// ---------------------------------------------------------------------------
void __fastcall TFMicroServer::btn_clearClick(TObject* Sender) {
	messagesLog->Lines->Clear();}
// ---------------------------------------------------------------------------

void __fastcall TFMicroServer::btn_startClick(TObject* Sender) {
	// ... START SERVER:

	// ... clear the Bindings property ( ... Socket Handles )
	IdTCPServer->Bindings->Clear();
	// ->->-> Bindings is a property of class: TIdSocketHandles;

	// ->->-> add listening ports:

	// ->->-> add a port for connections from guest clients->
	IdTCPServer->Bindings->Add()->Port = GUEST_CLIENT_PORT;
	// ->->-> etc->->

	// ->->-> ok, Active the Server!
	IdTCPServer->Active = true;
	// ->->-> disable start button
	btn_start->Enabled = false;

	// ->->-> enable stop button
	btn_stop->Enabled = true;

	// ->->-> message log
	Display("SERVER", "STARTED!");
}
// ---------------------------------------------------------------------------

void __fastcall TFMicroServer::btn_stopClick(TObject* Sender) {
	BroadcastMessage("Goodbye Client ");

	// ... stop server!
	IdTCPServer->Active = false;

	// ... hide stop button
	btn_stop->Enabled = false;

	// ... show start button
	btn_start->Enabled = true;

	// ... message log
	Display("SERVER", "STOPPED!");
}
// ---------------------------------------------------------------------------
